@@ -19,33 +19,60 @@ __docformat__ = 'restructuredtext'
 
 import sys
 
-from zope import component
-from zope.interface import implementer
-from zope.component.interfaces import ComponentLookupError
-from zope.traversing.interfaces import IPathAdapter, ITraversable
-from zope.traversing.interfaces import TraversalError
-from zope.traversing.adapters import traversePathElement
-from zope.security.proxy import ProxyFactory, removeSecurityProxy
 from zope.i18n import translate
+from zope.interface import implementer
+from zope.interface.interfaces import ComponentLookupError
+from zope.proxy import isProxy
+from zope.security.proxy import ProxyFactory
+from zope.security.proxy import removeSecurityProxy
+from zope.traversing.adapters import traversePathElement
+from zope.traversing.interfaces import IPathAdapter
+from zope.traversing.interfaces import ITraversable
+from zope.traversing.interfaces import TraversalError
+
+from zope import component
+
 
 try:
+    # The ``untrusted`` extra is needed to have zope.untrustedpython:
     from zope.untrustedpython import rcompile
     from zope.untrustedpython.builtins import SafeBuiltins
+
+    def guarded_getitem(ob, index):
+        """getitem access which gets guarded in the next line."""
+        return ob[index]
+    guarded_getitem = ProxyFactory(guarded_getitem)
     HAVE_UNTRUSTED = True
-except ImportError:
+except ImportError:   # pragma: no cover
     HAVE_UNTRUSTED = False
 
-from zope.tales.expressions import PathExpr, StringExpr, NotExpr, DeferExpr
+# PyPy doesn't support assigning to '__builtins__', even when using eval()
+# (http://pypy.readthedocs.org/en/latest/cpython_differences.html), so don't
+# try to use it. It won't work.
+if HAVE_UNTRUSTED:  # pragma: no cover
+    import platform
+    if platform.python_implementation() == 'PyPy':
+        HAVE_UNTRUSTED = False
+        del rcompile
+        del SafeBuiltins
+
+from zope.tales.expressions import DeferExpr
+from zope.tales.expressions import NotExpr
+from zope.tales.expressions import PathExpr
 from zope.tales.expressions import SimpleModuleImporter
+from zope.tales.expressions import StringExpr
 from zope.tales.pythonexpr import PythonExpr
-from zope.tales.tales import ExpressionEngine, Context
+from zope.tales.tales import Context
+from zope.tales.tales import ExpressionEngine
 
 from zope.pagetemplate.i18n import ZopeMessageFactory as _
 
+
 class InlineCodeError(Exception):
     pass
 
-class ZopeTraverser(object):
+
+class ZopeTraverser:
 
     def __init__(self, proxify=None):
         if proxify is None:
@@ -66,26 +93,31 @@ class ZopeTraverser(object):
             # special-case dicts for performance reasons
             if getattr(object, '__class__', None) == dict:
                 object = object[name]
+            elif isinstance(object, dict) and not isProxy(object):
+                object = object[name]
             else:
                 object = traversePathElement(object, name, path_items,
                                              request=request)
             object = self.proxify(object)
         return object
 
+
 zopeTraverser = ZopeTraverser(ProxyFactory)
 
+
 class ZopePathExpr(PathExpr):
 
     def __init__(self, name, expr, engine):
-        super(ZopePathExpr, self).__init__(name, expr, engine, zopeTraverser)
+        super().__init__(name, expr, engine, zopeTraverser)
+
 
 trustedZopeTraverser = ZopeTraverser()
 
+
 class TrustedZopePathExpr(PathExpr):
 
     def __init__(self, name, expr, engine):
-        super(TrustedZopePathExpr, self).__init__(name, expr, engine,
-                                                  trustedZopeTraverser)
+        super().__init__(name, expr, engine, trustedZopeTraverser)
 
 
 # Create a version of the restricted built-ins that uses a safe
@@ -100,15 +132,24 @@ class ZopePythonExpr(PythonExpr):
         def __call__(self, econtext):
             __traceback_info__ = self.text
             vars = self._bind_used_names(econtext, SafeBuiltins)
+            vars['_getattr_'] = SafeBuiltins.getattr
+            vars['_getitem_'] = guarded_getitem
             return eval(self._code, vars)
 
         def _compile(self, text, filename):
             return rcompile.compile(text, filename, 'eval')
 
 
+def _get_iinterpreter():
+    from zope.app.interpreter.interfaces import IInterpreter
+    return IInterpreter  # pragma: no cover
+
+
 class ZopeContextBase(Context):
     """Base class for both trusted and untrusted evaluation contexts."""
 
+    request = None
+
     def translate(self, msgid, domain=None, mapping=None, default=None):
         return translate(msgid, domain, mapping,
                          context=self.request, default=default)
@@ -118,24 +159,24 @@ class ZopeContextBase(Context):
     def evaluateCode(self, lang, code):
         if not self.evaluateInlineCode:
             raise InlineCodeError(
-                  _('Inline Code Evaluation is deactivated, which means that '
-                    'you cannot have inline code snippets in your Page '
-                    'Template. Activate Inline Code Evaluation and try again.'))
+                _('Inline Code Evaluation is deactivated, which means that '
+                  'you cannot have inline code snippets in your Page '
+                  'Template. Activate Inline Code Evaluation and try again.'))
 
         # TODO This is only needed when self.evaluateInlineCode is true,
         # so should only be needed for zope.app.pythonpage.
-        from zope.app.interpreter.interfaces import IInterpreter
+        IInterpreter = _get_iinterpreter()
         interpreter = component.queryUtility(IInterpreter, lang)
         if interpreter is None:
             error = _('No interpreter named "${lang_name}" was found.',
                       mapping={'lang_name': lang})
             raise InlineCodeError(error)
 
-        globals = self.vars.copy()
-        result = interpreter.evaluateRawCode(code, globals)
+        globs = self.vars.copy()
+        result = interpreter.evaluateRawCode(code, globs)
         # Add possibly new global variables.
         old_names = self.vars.keys()
-        for name, value in globals.items():
+        for name, value in globs.items():
             if name not in old_names:
                 self.setGlobal(name, value)
         return result
@@ -152,14 +193,20 @@ class ZopeContext(ZopeContextBase):
         Therefore, this method removes any proxy from the evaluated
         expression.
 
-        >>> output = [('version', 'xxx'), ('mode', 'html'), ('other', 'things')]
+        >>> from zope.pagetemplate.engine import ZopeContext
+        >>> from zope.tales.tales import ExpressionEngine
+        >>> from zope.security.proxy import ProxyFactory
+        >>> output = [
+        ...     ('version', 'xxx'),
+        ...     ('mode', 'html'),
+        ...     ('other', 'things')]
         >>> def expression(context):
         ...     return ProxyFactory(output)
         ...
         >>> zc = ZopeContext(ExpressionEngine, {})
         >>> out = zc.evaluateMacro(expression)
-        >>> type(out)
-        <type 'list'>
+        >>> type(out) is list
+        True
 
         The method does some trivial checking to make sure we are getting
         back a macro like we expect: it must be a sequence of sequences, in
@@ -204,7 +251,7 @@ class TrustedZopeContext(ZopeContextBase):
     """Evaluation context for trusted programs."""
 
 
-class AdapterNamespaces(object):
+class AdapterNamespaces:
     """Simulate tales function namespaces with adapter lookup.
 
     When we are asked for a namespace, we return an object that
@@ -218,11 +265,13 @@ class AdapterNamespaces(object):
       >>> def adapter1(ob):
       ...     return 1
       >>> adapter1.__component_adapts__ = (None,)
+      >>> from zope.traversing.interfaces import IPathAdapter
       >>> provideAdapter(adapter1, None, IPathAdapter, 'a1')
 
     Now, with this adapter in place, we can try out the namespaces:
 
       >>> ob = object()
+      >>> from zope.pagetemplate.engine import AdapterNamespaces
       >>> namespaces = AdapterNamespaces()
       >>> namespace = namespaces['a1']
       >>> namespace(ob)
@@ -282,12 +331,15 @@ class ZopeBaseEngine(ExpressionEngine):
 
         return context
 
+
 class ZopeEngine(ZopeBaseEngine):
-    """Untrusted expression engine.
+    """
+    Untrusted expression engine.
 
     This engine does not allow modules to be imported; only modules
     already available may be accessed::
 
+      >>> from zope.pagetemplate.engine import _Engine
       >>> modname = 'zope.pagetemplate.tests.trusted'
       >>> engine = _Engine()
       >>> context = engine.getContext(engine.getBaseNames())
@@ -311,20 +363,21 @@ class ZopeEngine(ZopeBaseEngine):
       >>> m._getframe
       Traceback (most recent call last):
         ...
-      ForbiddenAttribute: ('_getframe', <module 'sys' (built-in)>)
+      zope.security.interfaces.ForbiddenAttribute: ('_getframe', <module 'sys' (built-in)>)
 
     The results of Python expressions evaluated by this engine are
     wrapped in security proxies if the 'untrusted' extra is installed::
 
       >>> r = context.evaluate('python: {12: object()}.values')
-      >>> str(type(r).__name__) == (
-      ...   '_Proxy' if HAVE_UNTRUSTED else
-      ...   'builtin_function_or_method')
+      >>> str(type(r).__name__) if HAVE_UNTRUSTED else '_Proxy'
+      '_Proxy'
+      >>> ((str(type(r).__name__) in ('method', 'instancemethod'))
+      ... if not HAVE_UNTRUSTED else True)
       True
 
-      >>> r = context.evaluate('python: {12: object()}[12].__class__')
-      >>> str(type(r).__name__) == '_Proxy' or not HAVE_UNTRUSTED
-      True
+      >>> r = context.evaluate('python: {12: (1, 2, 3)}[12]')
+      >>> str(type(r).__name__) if HAVE_UNTRUSTED else '_Proxy'
+      '_Proxy'
 
     General path expressions provide objects that are wrapped in
     security proxies as well::
@@ -355,16 +408,16 @@ class ZopeEngine(ZopeBaseEngine):
       >>> o1.__name__
       'bar'
       >>> type(o1)
-      <type 'zope.security._proxy._Proxy'>
+      <class 'zope.security._proxy._Proxy'>
 
       >>> o2 = context.evaluate('foo/bar/baz')
       >>> o2.__name__
       'baz'
       >>> type(o2)
-      <type 'zope.security._proxy._Proxy'>
+      <class 'zope.security._proxy._Proxy'>
       >>> o3 = o2.__parent__
       >>> type(o3)
-      <type 'zope.security._proxy._Proxy'>
+      <class 'zope.security._proxy._Proxy'>
       >>> o1 == o3
       True
 
@@ -373,7 +426,7 @@ class ZopeEngine(ZopeBaseEngine):
 
     Note that this engine special-cases dicts during path traversal:
     it traverses only to their items, but not to their attributes
-    (e.g. methods on dicts), because of performance reasons:
+    (e.g. methods on dicts), because of performance reasons::
 
       >>> d = engine.getBaseNames()
       >>> d['adict'] = {'items': 123}
@@ -386,20 +439,42 @@ class ZopeEngine(ZopeBaseEngine):
         ...
       KeyError: 'keys'
 
+    This special-casing also applies to non-proxied dict subclasses::
+
+      >>> class TraverserDict(dict):
+      ...     def __init__(self):
+      ...         self.item_requested = None
+      ...     def __getitem__(self, item):
+      ...         self.item_requested = item
+      ...         return dict.__getitem__(self, item)
+
+      >>> d = engine.getBaseNames()
+      >>> foo = TraverserDict()
+      >>> d['foo'] = foo
+      >>> foo['bar'] = 'baz'
+      >>> context = engine.getContext(d)
+      >>> context.evaluate('foo/bar')
+      'baz'
+      >>> foo.item_requested
+      'bar'
+
       >>> tearDown()
 
-    """
+    """  # noqa: E501 line too long
 
     def getFunctionNamespace(self, namespacename):
         """ Returns the function namespace """
         return ProxyFactory(
-            super(ZopeEngine, self).getFunctionNamespace(namespacename))
+            super().getFunctionNamespace(namespacename))
+
 
 class TrustedZopeEngine(ZopeBaseEngine):
-    """Trusted expression engine.
+    """
+    Trusted expression engine.
 
     This engine allows modules to be imported::
 
+      >>> from zope.pagetemplate.engine import _TrustedEngine
       >>> modname = 'zope.pagetemplate.tests.trusted'
       >>> engine = _TrustedEngine()
       >>> context = engine.getContext(engine.getBaseNames())
@@ -413,7 +488,7 @@ class TrustedZopeEngine(ZopeBaseEngine):
       True
 
     Since this is trusted code, we can look at whatever is in the
-    module, not just __name__ or what's declared in a security
+    module, not just ``__name__`` or what's declared in a security
     assertion::
 
       >>> m.x
@@ -434,7 +509,7 @@ class TraversableModuleImporter(SimpleModuleImporter):
     def traverse(self, name, further_path):
         try:
             return self[name]
-        except KeyError:
+        except ImportError:
             raise TraversalError(self, name)
 
 
@@ -452,6 +527,7 @@ def _Engine(engine=None):
 
     return engine
 
+
 def _TrustedEngine(engine=None):
     if engine is None:
         engine = TrustedZopeEngine()
@@ -460,6 +536,7 @@ def _TrustedEngine(engine=None):
     engine.registerBaseName('modules', TraversableModuleImporter())
     return engine
 
+
 def _create_base_engine(engine, pathtype):
     for pt in pathtype._default_type_names:
         engine.registerType(pt, pathtype)
@@ -473,13 +550,13 @@ Engine = _Engine()
 TrustedEngine = _TrustedEngine()
 
 
-class AppPT(object):
+class AppPT:
 
     def pt_getEngine(self):
         return Engine
 
 
-class TrustedAppPT(object):
+class TrustedAppPT:
 
     def pt_getEngine(self):
         return TrustedEngine
